indexing
	description: "[
					  A genealogical database,
					  represented through a single, flat interface

					  Project 2 of the Software Engineering course 2008
					  ]"
	author: "Bernd Schoeller <bernd.schoeller@inf.ethz.ch>"
	date: "$Date: 2008-05-20 21:01:05 +0200 (Tue, 20 May 2008) $"
	revision: "$Revision: 14 $"

class
	GENEALOGY1

-- REMARKS:

-- Dates are represented as STRINGs. They use the form "TT.MM.YYYY",
-- where TT is the day using two digits, MM is the month using two
-- digits and YYYY is the year using four digits. ?? (year: ????) can
-- be used to represent unknown parts of the date. If the year is
-- unknown, the month needs to be unknown as well. If the month is
-- unknown, the day needs to be unknown as well.

-- Validity of all data entered should be sound, after every
-- operation. This must be enforced by the underlying implementation.
-- Your implementation must reject any operation that would make the
-- database unsound.

-- The data needs to be chronologically sound: a person cannot die
-- before he is born, or a father be younger than his child. It is not
-- necessary to model natural limits like an upper age limit or a
-- minimum or maximum age for child birth.

-- Marriages are not terminated by the death of one partner, but the
-- widow can marry another person without the need for a divorce -
-- this implicitly terminates the old marriage.

-- The whole interface implements "defensive programming", that means
-- that no extra preconditions should be added, but instead invalid
-- operations should report errors using the
-- 'last_operation_successful' and 'last_error' reporting
-- facilities. You are allowed (and encouraged) to add your
-- postconditions and invariants.

create
	make_empty

feature {NONE} -- Initialization

	make_empty is
			-- Make an empty genealogical database.
		do
			create {ARRAYED_LIST[PERSON1]} m_persons.make (0)
		    create {STRING} m_last_error.make_empty
		end

feature {NONE} -- Implementation

	m_last_error: STRING
	
	m_persons: LIST[PERSON1]

	set_error(error_message: STRING) is
			-- Sets the error message for error reporting
			require
				not_void: error_message /= void
				not_empty: not error_message.is_empty
			do
				m_last_error := error_message
			end

	clear_error is
			-- Clears the last error when an operation succeded
			require
				not_void: m_last_error /= void
			do
				m_last_error.clear_all
			end

	find_person (a_person_id : INTEGER) : PERSON1 is
			-- finds a person by the id. Void if no match is found
			do
				Result := void
				from
					m_persons.start
				until
					m_persons.after or Result /= void
				loop
					if m_persons.item.person_id = a_person_id then
						Result := m_persons.item
					end
					m_persons.forth
				end
			end

feature -- Error reporting

	last_operation_successful: BOOLEAN is
			-- Was the last operation successful?
			do
				Result := m_last_error.is_empty
			end

	last_error: STRING is
			-- Reason for the last failed operation (single line string)
			require
				error_available: not last_operation_successful
			do
				Result := m_last_error
			ensure
				not_void: Result /= Void
			end

feature -- Individuals changes

	add_male (an_id: INTEGER; a_first_name, a_last_name, a_date: STRING) is
			-- Add a new male person with id `an_id', first name
			-- `a_first_name', last name `a_last_name' and birthday
			-- `a_date'.
			require
				id_positive: an_id > 0
				first_name_not_void: a_first_name /= Void
				last_name_not_void: a_last_name /= Void
				date_valid: valid_date (a_date)
				fresh_id: not has (an_id)
			do
				clear_error
				m_persons.extend (create{PERSON1}.make (an_id, a_first_name, a_last_name, a_date, true))
			end

	add_female (an_id: INTEGER; a_first_name, a_last_name, a_date: STRING) is
			-- Add a new male person with id `an_id', first name
			-- `a_first_name', last name `a_last_name' and birthday
			-- `a_date'.
			require
				id_positive: an_id > 0
				first_name_not_void: a_first_name /= Void
				last_name_not_void: a_last_name /= Void
				date_valid: valid_date (a_date)
				fresh_id: not has (an_id)
			do
				clear_error
				m_persons.extend (create{PERSON1}.make (an_id, a_first_name, a_last_name, a_date, false))
			end

	register_death (an_id: INTEGER; a_date: STRING) is
			-- Register the death of person with `an_id', dying on `a_date'.
			require
				date_valid: valid_date (a_date)
				valid_person: has (an_id)
			local
				person: PERSON1
				date: GEN_DATE
			do
				-- Prepare
				clear_error
				create{GEN_DATE} date.make_from_string (a_date)
				person := find_person (an_id)

				-- Check if person is already dead
				if last_operation_successful and person.is_dead then
					set_error ("The person is already dead")
				end

				-- Check chronological order
				if last_operation_successful and person.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + person.last_action)
				end

				-- Set death
				if last_operation_successful then
					person.register_death (a_date)
				end
			end

feature -- Individual status

	has (an_id: INTEGER): BOOLEAN is
			-- Is there a person with id `an_id'?
			do
				clear_error
				Result := (find_person (an_id) /= void)
			ensure
				only_valid_ids_exist: (an_id <= 0) implies not Result
			end

	is_male (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' male?
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).is_male
			end

	is_female (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' female?
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).is_female
			end

	is_alive (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' alive?
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).is_alive
			end

	is_dead (an_id: INTEGER): BOOLEAN is
			-- Is person `an_id' dead?
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).is_dead
			end

feature -- Individual query

	first_name (an_id: INTEGER): STRING is
			-- First name of `an_id'
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).first_name
			end

	last_name (an_id: INTEGER): STRING is
			-- (Current) last name of `an_id'
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).last_name
			end

	birth_name (an_id: INTEGER): STRING is
			-- Birth name of `an_id'
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).birth_name
			end

	birth_date (an_id: INTEGER): STRING is
			-- Birth date of `an_id'
			require
				valid_person: has (an_id)
			do
				clear_error
				Result := find_person (an_id).birth_date.date_string
			end

	death_date (an_id: INTEGER): STRING is
			-- Death date of `an_id', Void if the person is not dead
			require
				valid_person: has (an_id)
			local
				person : PERSON1
			do
				clear_error
				person := find_person (an_id)
				if person.is_alive then
					Result := void
				else
					Result := person.death_date.date_string
				end
			end

	find (a_name: STRING): LIST [INTEGER] is
			-- Persons with first or last name equal to `a_name'
			require
				not_void: a_name /= Void
			do
				clear_error
				create {ARRAYED_LIST[INTEGER]} Result.make (0)
				from
					m_persons.start
				until
					m_persons.after
				loop
					if m_persons.item.first_name = a_name or m_persons.item.last_name = a_name then
						Result.extend (m_persons.item.person_id)
					end
					m_persons.forth
				end
			ensure
				not_void: Result /= Void
			end

feature -- Relations

	register_child (a_parent, a_child: INTEGER) is
			-- Register `a_child' as the (natural) child of `a_parent',
			-- making `a_parent' mother or father of `a_child'.
			require
				valid_parent: has (a_parent)
				valid_child: has (a_child)
			local
				parent, child: PERSON1
				date : GEN_DATE
			do
				-- Prepare
				clear_error
				parent := find_person (a_parent)
				child := find_person (a_child)

				-- Check if parent and child are dead
				if last_operation_successful and (parent.is_dead or child.is_dead) then
					set_error ("Parent and/or child is dead")
				end

				-- Check if child and Parent are the same person
				if last_operation_successful and a_parent = a_child then
					set_error ("Parent and child are the same person")
				end

				-- Check if natural parent already exists
				if last_operation_successful and child.has_natural_parent (parent.is_male) then
					set_error ("The children has already a natural parent with this sex")
				end

				-- Check if parent is ancestor of child
				if last_operation_successful and parent.has_ancestor (child) then
					set_error ("The parent is ancestor of the child")
				end

				-- Check if child is ancestor of parent
				if last_operation_successful and child.has_ancestor (parent) then
					set_error ("The child is ancestor of the parent")
				end

				-- find youngest ancestor of the parent and check the dates.
				-- The youngest ancestor must not be younger than the child
				if last_operation_successful then
					date := parent.birth_date
					from
						m_persons.start
					until
						m_persons.after
					loop
						if parent.has_ancestor (m_persons.item) then
							if date.is_undefined or m_persons.item.birth_date.is_newer_date (date) then
								date := m_persons.item.birth_date
							end
						end
						m_persons.forth
					end

					if date.is_newer_date (child.birth_date) then
						set_error ("The parent or one of the ancestors are younger then the child")
					end
				end

				-- Update parent and child
				if last_operation_successful then
					child.add_parent (parent)
					parent.add_child (child)
				end
			end

	adopt (a_parent, a_child: INTEGER; a_last_name: STRING; a_date: STRING) is
			-- Adopt `a_child' by `a_parent' on `a_date', changing the
			-- last name of the child to `a_last_name'.
			require
				valid_parent: has (a_parent)
				valid_child: has (a_child)
				valid_date: valid_date (a_date)
			local
				parent, child : PERSON1
				date : GEN_DATE
			do
				-- Prepare
				clear_error
				parent := find_person (a_parent)
				child := find_person (a_child)
				create {GEN_DATE} date.make_from_string (a_date)

				-- Check if parent and child are the same
				if last_operation_successful and a_parent = a_child then
					set_error ("Parent and child are the same person")
				end

				-- Check if parent and child are alive
				if last_operation_successful and (child.is_dead or parent.is_dead) then
					set_error ("The child and/or the parent are dead")
				end

				-- Check if child is ancestor
				if last_operation_successful and parent.has_ancestor (child) then
					set_error ("The child is an ancestor of the parent")
				end

				-- Check chronological order of the parent
				if last_operation_successful and parent.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + parent.last_action)
				end

				-- Check chronological order of the child
				if last_operation_successful and child.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + child.last_action)
				end

				-- Check if child has already an adoptive parent
				if last_operation_successful and child.has_living_adoptive_parent (parent.is_male) then
					set_error ("Child has already an adoptive parent of this sex")
				end

				-- Check for direct relatives
				if last_operation_successful and child.is_direct_relative (parent) then
					set_error ("Child and parent are direct relatives")
				end

				-- Everything OK => Adopt
				if last_operation_successful then
					parent.adopt_child (child, a_date)
					child.adopt_parent (parent, a_date, a_last_name)
				end
			end

feature -- Relation status

	is_natural_father (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' the natural father of `a_child'?
			require
				valid_parent: has (a_parent)
				valid_child: has (a_child)
			local
				parent, child : PERSON1
			do
				clear_error
				parent := find_person (a_parent)
				child := find_person (a_child)
				Result := parent.is_male and child.natural_parents.has (parent)
			end

	is_natural_mother (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' the natural mother of `a_child'?
			require
				valid_parent: has (a_parent)
				valid_child: has (a_child)
			local
				parent, child : PERSON1
			do
				clear_error
				parent := find_person (a_parent)
				child := find_person (a_child)
				Result := parent.is_female and child.natural_parents.has (parent)
			end

	is_parent (a_parent, a_child: INTEGER): BOOLEAN is
			-- Is `a_parent' a parent (natural or adopted) of `a_child'?
			require
				valid_parent: has (a_parent)
				valid_child: has (a_child)
			do
			    Result := parents (a_child).has (a_parent)
			end

feature -- Relation queries

	father (a_child: INTEGER): INTEGER is
			-- The (current) father of `a_child'
			require
				valid_child: has (a_child)
			local
				child, current_father : PERSON1
			do
				clear_error
				child := find_person (a_child)
				current_father := child.current_parent (true)
				if current_father /= void then
					Result := current_father.person_id
				end
			end

	mother (a_child: INTEGER): INTEGER is
			-- The (current) mother of `a_child'
			require
				valid_child: has (a_child)
			local
				child, current_mother : PERSON1
			do
				clear_error
				child := find_person (a_child)
				current_mother := child.current_parent (false)
				if current_mother /= void then
					Result := current_mother.person_id
				end
			end

	parents (a_child: INTEGER): LIST [INTEGER] is
			-- List of all parents of `a_child', including adoptions
			require
				valid_child: has (a_child)
			do
				Result := find_person (a_child).all_parents
			ensure
				not_void: Result /= Void
				father_included: (father (a_child) > 0) implies
									  Result.has (father (a_child))
				mother_included: (mother (a_child) > 0) implies
									  Result.has (mother (a_child))
			end


	children (a_parent: INTEGER): LIST [INTEGER] is
			-- All children (natural or adopted) of `a_parent'
			require
				valid_parent: has (a_parent)
			do
				Result := find_person (a_parent).all_children
			ensure
				not_void: Result /= Void
			end

feature -- Marriages

	marry (a_first_person,a_second_person: INTEGER;
			 a_first_last_name, a_second_last_name: STRING;
			 a_date: STRING) is
			-- `a_first_person' marries `a_second_person' on `a_date',
			-- resulting in the first person to change the last name to
			-- `a_first_last_name' (Void = no change) and the second
			-- person to change the last name to `a_second_last_name'
			-- (Void = no change).
			require
				valid_first: has (a_first_person)
				valid_second: has (a_second_person)
				valid_date: valid_date (a_date)
			local
				first_person, second_person : PERSON1
				date : GEN_DATE
			do
				-- Prepare
				clear_error
				first_person := find_person (a_first_person)
				second_person := find_person (a_second_person)
				create{GEN_DATE} date.make_from_string (a_date)

				-- Check if parent and child are alive
				if last_operation_successful and (first_person.is_dead or second_person.is_dead) then
					set_error ("The first and/or second person is dead")
				end

				-- Check if child and Parent are the same person
				if last_operation_successful and a_first_person = a_second_person then
					set_error ("The two persons are the same")
				end

				-- Check if the first person is married
				if last_operation_successful and first_person.is_married and first_person.current_partner.person.is_alive then
					set_error ("First person is already married")
				end

				-- Check if the second person is married
				if last_operation_successful and second_person.is_married and second_person.current_partner.person.is_alive then
					set_error ("The second person is already married")
				end

				-- Check next relatives
				if last_operation_successful and second_person.is_direct_relative (first_person) then
					set_error ("The two persons are direct relatives")
				end

				-- Check chronological order for the first person
				if last_operation_successful and first_person.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + first_person.last_action)
				end

				-- Check chronological order for the second person
				if last_operation_successful and second_person.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + second_person.last_action)
				end

				-- Everything OK => marry
				if last_operation_successful then
					first_person.marry (second_person, a_date, a_first_last_name)
					second_person.marry (first_person, a_date, a_second_last_name)
				end
			end

	divorce (an_id: INTEGER; a_first_last_name, other_last_name: STRING;
				a_date: STRING) is
			-- `an_id' divorces from the current partner on `a_date',
			-- resulting in the person to change the last name to
			-- `a_last_name' (Void = no change) and the spouse to change
			-- the last name to `other_last_name' (Void = no change).
			require
				valid_id: has (an_id)
				valid_date (a_date)
			local
				person : PERSON1
				date : GEN_DATE
			do
				-- Prepare
				clear_error
				create{GEN_DATE} date.make_from_string (a_date)
				person := find_person (an_id)

				-- Check if person is married
				if last_operation_successful and not person.is_married then
					set_error ("Person is not married")
				end

				-- Check if person or parnter is dead
				if last_operation_successful and (person.is_dead or person.current_partner.person.is_dead) then
					set_error ("Person and/or partner is dead")
				end

				-- Check chronological order for the person
				if last_operation_successful and person.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + person.last_action)
				end

				-- Check chronological order for the partner
				if last_operation_successful and person.current_partner.person.last_date.is_newer_date (date) then
					set_error ("Cronological error: " + person.current_partner.person.last_action)
				end

				-- Everything OK => Divorce
				if last_operation_successful then
					person.current_partner.person.divorce (a_date, other_last_name)
					person.divorce (a_date, a_first_last_name)
				end
			end

feature -- Marriage status

	is_married (an_id: INTEGER): BOOLEAN is
			-- Is `an_id' currently married?
			require
				valid_id: has (an_id)
			do
				clear_error
				Result := find_person (an_id).is_married
			end

feature -- Marriage queries

	partners (an_id: INTEGER): LIST [INTEGER] is
			-- Ordered list of all partners of `an_id'. Duplicates are
			-- possible if the same person was married multiple times.
			require
				valid_id: has (an_id)
			do
				Result := find_person (an_id).all_partner
			ensure
				not_void: Result /= Void
			end

	marriage_dates (an_id: INTEGER): LIST [STRING] is
			-- Date of all marriage dates of `an_id', corresponding to
			-- the `partners' query
			require
				valid_id: has (an_id)
			do
				Result := find_person (an_id).all_marriage_dates
			ensure
				not_void: Result /= Void
				same_length: Result.count = partners (an_id).count
			end

	divorce_dates (an_id: INTEGER): LIST [STRING] is
			-- Date of all divorce dates of `an_id', corresponding to the
			-- `partners' query; Void if the corresponding marriage was
			-- not divorced
			require
				valid_id: has (an_id)
			do
				Result := find_person (an_id).all_divorce_dates
			ensure
				not_void: Result /= Void
				same_length: Result.count = partners (an_id).count
			end

feature -- Analysis status

	is_ancestor (a_first_id, a_second_id: INTEGER): BOOLEAN is
			-- Is `a_first_id' an ancestor of `a_second_id'?
			-- (including adoptions)
			require
				valid_first: has (a_first_id)
				valid_second: has (a_second_id)
			local
				first_person, second_person : PERSON1
			do
				clear_error

				-- Check if same person
				if a_first_id = a_second_id then
					set_error ("First and second are the same person")
				end

				if last_operation_successful then
					first_person := find_person (a_first_id)
					second_person := find_person (a_second_id)
					Result := second_person.has_ancestor (first_person)
				end
			end

feature -- Analysis queries

	closest_common_ancestor (a_first_id, a_second_id: INTEGER): INTEGER is
			-- The id of the closest common ancestor between `a_first_id'
			-- and `a_second_id'. If there is a common ancestor only
			-- following natural relationships, this one should be
			-- preferred.  Otherwise, the number of adoptions should be
			-- minimized. If there is more than one closest ancestor, the
			-- result is non-deterministic. 0 if there is none.
			require
				valid_first: has (a_first_id)
				valid_second: has (a_second_id)
			local
				finder : RELATION_FINDER
				first, second : PERSON1
				relation : LIST[INTEGER]
			do
				-- Prepare
				clear_error
				first := find_person (a_first_id)
				second := find_person (a_second_id)
				Result := 0

				-- Check
				if a_first_id = a_second_id then
					set_error ("First and Second are the same person")
				end

				-- Check if Second is ancestor of First
				if last_operation_successful and Result = 0 and first.has_ancestor (second) then
					Result := second.person_id
				end

				-- Check if First is ancestor of Second
				if last_operation_successful and Result = 0 and second.has_ancestor (first) then
					Result := first.person_id
				end

				-- Configure Finder to find a relation between first and second:
				-- Disallow marriage relations
				-- Penalize adoption with a higher weight of 100
				if last_operation_successful and Result = 0 then
					create {RELATION_FINDER} finder.make (m_persons, first, second)
					finder.set_marriage_weight (-1)
					finder.set_natural_parent_weight (1)
					finder.set_adoption_weight (100)
					relation := finder.find_relation

					-- Find the node that is the common ancestor in the relation
					if relation /= void and relation.count > 0 then
						from
							relation.start
						until
							relation.after or Result > 0
						loop
							if is_ancestor (relation.item, a_first_id) and is_ancestor (relation.item, a_second_id) then
								Result := relation.item
							end
							relation.forth
						end
					end
				end
			end

	shortest_relation (a_first_id, a_second_id: INTEGER): LIST [INTEGER] is
			-- The shortest relation between `a_first_id' and
			-- `a_second_id', following child/parent, adoption and (old
			-- and current) marriage relations. The first element in the
			-- list is `a_first_id', the last element in the list is
			-- `a_second_id'. Void if there is no relation between the
			-- two.
			require
				valid_first: has (a_first_id)
				valid_second: has (a_second_id)
			local
				finder : RELATION_FINDER
				first, last : PERSON1
			do
				-- Prepare
				clear_error
				first := find_person (a_first_id)
				last := find_person (a_second_id)

				-- Check if the persons are the same
				if a_first_id = a_second_id then
					set_error ("First and second are the same person")
				end

				-- Default configuration of finder: all relations are allowed with the same weight
				if last_operation_successful then
					create {RELATION_FINDER} finder.make (m_persons, first, last)
					Result := finder.find_relation
				end

			ensure
				first_first_if_found: (Result /= Void) implies
											 Result.first = a_first_id
				second_last_if_found: (Result /= Void) implies
											 Result.last = a_second_id
			end

	best_relation (a_first_id, a_second_id: INTEGER): LIST [INTEGER] is
			-- The best relation between `a_first_id' and `a_second_id',
			-- following child/parent, adoption and marriage
			-- relations. The number of marriage relations should be
			-- minimized. For the same number of marriages, the number
			-- adoptions should be minimized. For the same number of
			-- adoptions, list size should be minimized. Void if there is
			-- no relation between the two.
			require
				valid_first: has (a_first_id)
				valid_second: has (a_second_id)
			local
				finder : RELATION_FINDER
				first, last : PERSON1
			do
				-- Prepare
				clear_error
				first := find_person (a_first_id)
				last := find_person (a_second_id)

				-- Check if the persons are the same
				if a_first_id = a_second_id then
					set_error ("First and second are the same person")
				end

				-- Configure Finder to find a relation between first and second:
				-- Penalize marriages and adoptions with a higher weight of 100
				if last_operation_successful then
					create {RELATION_FINDER} finder.make (m_persons, first, last)
					finder.set_marriage_weight (100)
					finder.set_adoption_weight (100)
					Result := finder.find_relation
				end

			ensure
				first_first_if_found: (Result /= Void) implies
											 Result.first = a_first_id
				second_last_if_found: (Result /= Void) implies
											 Result.last = a_second_id
			end





feature -- Date validation

	valid_date (a_date: STRING): BOOLEAN is
			-- Is `a_date' a valid date?
			local
				date:GEN_DATE
			do
				if a_date = void then
					Result := false
				elseif a_date.count /= 10 then
					Result := false
				else
					create{GEN_DATE} date.make_from_string (a_date)
					Result := date.is_valid
				end
			ensure
				void_implies_false: (a_date = Void) implies not Result
				wrong_length_implies_false: (a_date.count /= 10) implies not Result
			end
end
