indexing
	description: "Objects that represents a person in the genealogical database"
	author: "Thomas Gubler and Wolfgang Schwaiger"
	date: "$Date: 2008-05-20 21:01:05 +0200 (Tue, 20 May 2008) $"
	revision: "$Revision: 14 $"

class
	PERSON1

create
	make

feature -- Initialisation

	make(a_person_id:INTEGER; a_first_name, a_last_name, a_birth_date: STRING; is_a_male:BOOLEAN) is
			-- Make a new person with this parameters
			require
				valid_id: a_person_id > 0
				valid_first_name: a_first_name /=void and a_first_name.count > 0
				valid_last_name: a_last_name /= void and a_last_name.count > 0
			do
				m_person_id := a_person_id
				m_first_name := a_first_name
				m_birth_name := a_last_name
				m_last_name := a_last_name
				m_is_male := is_a_male
			    create {GEN_DATE}  m_birth_date.make_from_string (a_birth_date)
			    create {GEN_DATE}  m_death_date.make_nulldate
				create {ARRAYED_LIST[PERSON1]} m_natural_parents.make (0)
				create {ARRAYED_LIST[PERSON1]} m_children.make (0)
				create {ARRAYED_LIST[MARRIAGE]} m_partners.make (0)
				create {ARRAYED_LIST[ADOPTION]} m_adoptive_parents.make (0)
				create {ARRAYED_LIST[ADOPTION]} m_adopted_children.make (0)

				-- Log
				create {GEN_DATE} m_last_date.make_nulldate
				log_action ("Birth", a_birth_date, void)

			end

feature{NONE} -- Implementation

	m_person_id: INTEGER
	m_first_name: STRING
	m_last_name: STRING
	m_birth_name: STRING
	m_birth_date: GEN_DATE
	m_death_date: GEN_DATE
	m_is_male: BOOLEAN

	m_children : LIST[PERSON1]
	m_natural_parents : LIST[PERSON1]

	m_partners: LIST[MARRIAGE]
	m_adoptive_parents : LIST[ADOPTION]
	m_adopted_children : LIST[ADOPTION]

	-- Relation finding
	m_runner : RUNNER
	m_distance : INTEGER


feature{NONE} -- Logging

	-- Log last action
	m_last_action : STRING
	m_last_date : GEN_DATE

	log_action (a_description, a_date : STRING; other_person : PERSON1) is
			-- Keeps trace of the newest action of the current person or the other_person
			-- having some relationship to the current person, in order to avoid chronological errors
			local
				date : GEN_DATE
			do
				create {GEN_DATE} date.make_from_string (a_date)

				-- Check if current action is newer then m_last_action
				if m_last_date.is_null_date or m_last_date.is_undefined or date.is_newer_date (m_last_date) then
					m_last_date.set_date (date.date_string)
					m_last_action := m_first_name + " " + m_last_name + ", " + a_description + " on " + m_last_date.date_string
				end

				-- Check if the other person has some newer action then m_last_action
				if other_person /= void and not other_person.last_date.is_null_date and not other_person.last_date.is_undefined then
					if m_last_date.is_undefined or other_person.last_date.is_newer_date (m_last_date) then
						m_last_date.set_date (other_person.last_date.date_string)
						m_last_action := other_person.last_action
					end
				end
			end


feature -- Access

	person_id: INTEGER is
			-- Gets the unique identifier of a person
			do
				Result := m_person_id
			end

	first_name: STRING is
			-- Gets the firstname
			do
				Result := m_first_name
			end

	last_name: STRING is
			-- Gets the current lastname
			do
				Result := m_last_name
			end

	birth_name: STRING is
			-- Gets the birthname
			do
				Result := m_birth_name
			end

	birth_date: GEN_DATE is
			-- Gets the birth date
			do
				Result := m_birth_date
			end

	death_date: GEN_DATE is
			-- Gets the death date
			do
				Result := m_death_date
			end

	natural_parents: LIST[PERSON1] is
			-- Gets the natural parents
			do
				Result := m_natural_parents
			end

	children : LIST[PERSON1] is
			-- Gets the children
			do
				Result := m_children
			end

	adoptive_parents : LIST[ADOPTION] is
			-- Gets the adoptive parents
			do
				Result := m_adoptive_parents
			end

	adopted_children : LIST[ADOPTION] is
			-- Gets the adopted children
			do
				Result := m_adopted_children
			end

	current_parent (is_parent_male : BOOLEAN) : PERSON1 is
			-- Gets the current father
			do
				Result := void

				-- Check adoptive parents
				from
					m_adoptive_parents.start
				until
					m_adoptive_parents.after or Result /= void
				loop
					if m_adoptive_parents.item.person.is_male = is_parent_male then
						Result := m_adoptive_parents.item.person
					end
					m_adoptive_parents.forth
				end

				-- Check natural parents
				if Result /= void then
					from
						m_natural_parents.start
					until
						m_natural_parents.after or Result /= void
					loop
						if m_natural_parents.item.is_male = is_parent_male then
							Result := m_natural_parents.item
						end
						m_natural_parents.forth
					end
				end
			end

	all_parents : LIST[INTEGER] is
			-- Gets all parents, including adoptions
			do
				create {ARRAYED_LIST[INTEGER]} Result.make(0)

				-- Natural Parents
				from
					m_natural_parents.start
				until
					m_natural_parents.after
				loop
					Result.extend(m_natural_parents.item.person_id)
					m_natural_parents.forth
				end

				-- Adoptive Parents
				from
					m_adoptive_parents.start
				until
					m_adoptive_parents.after
				loop
					Result.extend(m_adoptive_parents.item.person.person_id)
					m_adoptive_parents.forth
				end
			end

	all_children : LIST[INTEGER] is
			-- Gets all children
			do
				create {ARRAYED_LIST[INTEGER]} Result.make(0)
				-- Natural Children
				from
					m_children.start
				until
					m_children.after
				loop
					Result.extend(m_children.item.person_id)
					m_children.forth
				end

				-- Adoptive Parents
				from
					m_adopted_children.start
				until
					m_adopted_children.after
				loop
					Result.extend(m_adopted_children.item.person.person_id)
					m_adopted_children.forth
				end
			end

	all_partner : LIST[INTEGER] is
			-- Gets all partners
			do
				create {ARRAYED_LIST[INTEGER]} Result.make(0)
				from
					m_partners.start
				until
					m_partners.after
				loop
					Result.extend(m_partners.item.person.person_id)
					m_partners.forth
				end

			end

	all_marriage_dates : LIST[STRING] is
			-- Gets all Marriage dates
			do
				create {ARRAYED_LIST[STRING]} Result.make(0)
				from
					m_partners.start
				until
					m_partners.after
				loop
					Result.extend(m_partners.item.start_date.date_string)
					m_partners.forth
				end

			end

	all_divorce_dates : LIST[STRING] is
			-- Gets all Marriage dates
			do
				create {ARRAYED_LIST[STRING]} Result.make(0)
				from
					m_partners.start
				until
					m_partners.after
				loop
					if m_partners.item.end_date.is_null_date then
						Result.extend(void)
					else
						Result.extend (m_partners.item.end_date.date_string)
					end
					m_partners.forth
				end

			end


	current_partner : MARRIAGE is
			-- Gets the current partner. void if no partner exists
			do
				if m_partners.count > 0 then
					Result := m_partners[m_partners.count]
				end
			end

	last_action : STRING is
			-- Gets the last Change of this person
			do
				Result := m_last_action
			end

	last_date : GEN_DATE is
			-- Gets the date of the last change for this person
			do
				Result := m_last_date
			end


feature -- Status report

	is_male: BOOLEAN is
			-- is this person male?
			do
				Result := m_is_male
			end

	is_female: BOOLEAN is
			-- is this person a female?
			do
				Result := not is_male
			end

	is_alive: BOOLEAN is
			-- is this person alive ?
			do
				Result := m_death_date.is_null_date
			end

	is_dead: BOOLEAN is
			-- is this person dead ?
			do
				Result := not is_alive
			end

	has_natural_parent (is_parent_male : BOOLEAN) : BOOLEAN is
			-- Has Current a natural parent with the given sex ?
			do
				Result := false
				from
					m_natural_parents.start
				until
					Result or m_natural_parents.after
				loop
					Result := (m_natural_parents.item.is_male = is_parent_male)
					m_natural_parents.forth
				end
			end

	has_living_adoptive_parent (is_parent_male : BOOLEAN) : BOOLEAN is
			-- Has Current a living adoptive parent with the given sex?
			do
				Result := false
				from
					m_adoptive_parents.start
				until
					Result or m_adoptive_parents.after
				loop
					Result := (m_adoptive_parents.item.person.is_male = is_parent_male and m_adoptive_parents.item.person.is_alive)
					m_adoptive_parents.forth
				end

			end


	is_married : BOOLEAN is
			-- Is Current married
			do
				if m_partners.count > 0 then
					Result := m_partners[m_partners.count].end_date.is_null_date
				else
					Result := false
				end
			end

	is_direct_relative (other : PERSON1) : BOOLEAN is
			-- is other direct ancestor, or silblings
			local
				current_parents : LIST[INTEGER]
			do
				Result := false

				-- Check children
				if not Result and all_children.has (other.person_id) then
					Result := true
				end

				-- Check parents
				if not Result and all_parents.has (other.person_id) then
					Result := true
				end

				-- Check silblings (have same parents)
				current_parents := all_parents
				if not Result then
					from
						current_parents.start
					until
						current_parents.after or Result
					loop
						Result := other.all_parents.has (current_parents.item)
						current_parents.forth
					end
				end
			end

	has_ancestor (a_person : PERSON1) : BOOLEAN is
			-- is 'a_person' an ancestor of Current?
			do
				Result := false

				-- Check natural parents
				if not Result then
					Result := m_natural_parents.has (a_person)
				end

				-- Check adoptive parents
				if not Result then
					Result := is_adoptive_parent(a_person)
				end

				-- Recursive check of natural parents
				if not Result then
					from
						m_natural_parents.start
					until
						Result or m_natural_parents.after
					loop
						Result := m_natural_parents.item.has_ancestor (a_person)
						m_natural_parents.forth
					end
				end

				-- Recursive check of adoptive parents
				if not Result then
					from
						m_adoptive_parents.start
					until
						Result or m_adoptive_parents.after
					loop
						Result := m_adoptive_parents.item.person.has_ancestor(a_person)
						m_adoptive_parents.forth
					end
				end
			end

	is_adoptive_parent (a_person : PERSON1) : BOOLEAN is
			-- is 'a_person' an adoptive parent of Current
			do
				Result := false
				from
					m_adoptive_parents.start
				until
					Result or m_adoptive_parents.after
				loop
					Result := m_adoptive_parents.item.person.person_id = a_person.person_id
					m_adoptive_parents.forth
				end
			end

feature -- Changes

	register_death (a_date : STRING) is
			-- Sets the death date
			do
				m_death_date.set_date (a_date)
				log_action ("Death", a_date, void)
			end


	marry (a_partner : PERSON1; a_date, a_last_name : STRING) is
			-- Marries Current with partner
			do
				m_partners.extend (create{MARRIAGE}.make (a_partner, a_date))
				m_last_name := a_last_name
				log_action ("Marriage", a_date, a_partner)
			end

	divorce (a_date, a_last_name : STRING)  is
			-- Divorces from the current partner
			do
				current_partner.end_date.set_date (a_date)
				if a_last_name /= void then
					m_last_name := a_last_name
				end
				log_action ("Divirce", a_date, current_partner.person)
			end

	add_child (a_child : PERSON1) is
			-- Adds a child (natural)
			do
				m_children.extend (a_child)
				log_action ("New Child", a_child.birth_date.date_string, a_child)
			end

	add_parent (a_parent : PERSON1) is
			-- Adds a parent (natural)
			do
				m_natural_parents.extend (a_parent)
				log_action ("New parent", m_birth_date.date_string, a_parent)
			end

	adopt_child (a_child : PERSON1; a_date : STRING) is
			-- Adopt a child
			do
				m_adopted_children.extend (create{ADOPTION}.make (a_child, a_date) )
				log_action ("Adopt a Child", a_date, a_child)
			end

	adopt_parent (a_parent: PERSON1; a_date, a_last_name : STRING) is
			-- Adopt a parent
			do
				m_adoptive_parents.extend (create{ADOPTION}.make (a_parent, a_date))
				m_last_name := a_last_name
				log_action ("Adopted by a parent", a_date, a_parent)
			end

feature -- Find Relations
    has_runner : BOOLEAN is
    		-- True if there is a runner in this node
    		do
    			Result := m_runner /= void
    		end

	runner : RUNNER is
			-- Gets the runner
			do
				Result := m_runner
			end

	distance : INTEGER is
			-- Gets the distance saved in this node
			do
				Result := m_distance
			end

	init_runner(first_id : INTEGER) is
			-- Initializes in the node for a new search
			do
				if first_id = m_person_id then
					create {RUNNER} m_runner.make_empty
					m_runner.path.extend (m_person_id)
				else
					m_runner := void
				end
				m_distance := -1
			end


	set_runner(a_runner : RUNNER; dist_increment : INTEGER) is
			-- Sets the runner for this node. Discarded if the saved distance in the node
			-- is smaller then the distance of the runner
			do
				if (m_distance < 0 or (a_runner.distance + dist_increment) < m_distance) then
					m_runner := a_runner
					m_runner.increment_distance (dist_increment)
					m_runner.path.extend (m_person_id)

					m_distance := m_runner.distance
				end
			end

	move_runner (natural_weight, adoption_weight, marriage_weight : INTEGER) is
			-- Move the runner to the next nodes
			do
				-- Natural Parants/Children
				if natural_weight > 0 then
					-- Parents
					from
						m_natural_parents.start
					until
						m_natural_parents.after
					loop
						m_natural_parents.item.set_runner (runner.deep_twin, natural_weight)
						m_natural_parents.forth
					end

					-- Children
					from
						m_children.start
					until
						m_children.after
					loop
						m_children.item.set_runner (m_runner.deep_twin, natural_weight)
						m_children.forth
					end
				end

				-- Adoptive Parents/Children
				if adoption_weight > 0 then
					-- Parents
					from
						m_adoptive_parents.start
					until
						m_adoptive_parents.after
					loop
						m_adoptive_parents.item.person.set_runner (m_runner.deep_twin, adoption_weight)
						m_adoptive_parents.forth
					end


					-- Children
					from
						m_adopted_children.start
					until
						m_adopted_children.after
					loop
						m_adopted_children.item.person.set_runner (m_runner.deep_twin, adoption_weight)
						m_adopted_children.forth
					end
				end

				-- Partners
				if marriage_weight > 0 then
					from
						m_partners.start
					until
						m_partners.after
					loop
						m_partners.item.person.set_runner (m_runner.deep_twin, marriage_weight)
						m_partners.forth
					end
				end

				-- Remove the runner from this node
				m_runner := void
			end

invariant
	invariant_clause: True -- Your invariant here
end
